/*
 *	PaperJam -- Command parser
 *
 *	(c) 2018 Martin Mares <mj@ucw.cz>
 */

#include <cassert>
#include <cstdarg>
#include <cstdlib>
#include <cstdio>

#include "jam.h"

/*** Lexer ***/

enum token_type {
  TOK_NONE,
  TOK_END,
  TOK_EQUAL,
  TOK_COMMA,
  TOK_COLON,
  TOK_DOTDOT,
  TOK_OPEN_PAREN,
  TOK_CLOSE_PAREN,
  TOK_OPEN_BRACE,
  TOK_CLOSE_BRACE,
  TOK_IDENT,
  TOK_STRING,
  TOK_NUMBER,
};

const char *in_pos;
static token_type this_token = TOK_NONE;
static token_type buffered_token = TOK_NONE;
static string token;
static double token_num;

static void parse_commands(list<cmd *> &cmds);

static token_type get_next_token()
{
  while (*in_pos == ' ' || *in_pos == '\t' || *in_pos == '\r' || *in_pos == '\n')
    in_pos++;

  token = "";
  if (!*in_pos)
    return TOK_END;

  if (*in_pos >= '0' && *in_pos <= '9' ||
      *in_pos == '-' && in_pos[1] >= '0' && in_pos[1] <= '9')
    {
      token += *in_pos++;
      while (*in_pos >= '0' && *in_pos <= '9' || *in_pos == '.' && in_pos[1] != '.')
	token += *in_pos++;

      size_t end_pos;
      token_num = stod(token, &end_pos);
      if (end_pos < token.length())
	err("Invalid number %s", token.c_str());
      return TOK_NUMBER;
    }

  if (*in_pos >= 'A' && *in_pos <= 'Z' ||
      *in_pos >= 'a' && *in_pos <= 'z')
    {
      while (*in_pos >= 'A' && *in_pos <= 'Z' ||
	     *in_pos >= 'a' && *in_pos <= 'z' ||
	     *in_pos >= '0' && *in_pos <= '9' ||
	     *in_pos == '_' ||
	     *in_pos == '-')
	token += *in_pos++;
      return TOK_IDENT;
    }

  if (*in_pos == '"')
    {
      in_pos++;
      while (*in_pos != '"')
	{
	  if (!*in_pos)
	    err("Unterminated string");
	  if (*in_pos == '\\')
	    {
	      in_pos++;
	      if (*in_pos != '"' && *in_pos != '\\')
		err("Unrecognized escape sequence \\%c", *in_pos);
	    }
	  token += *in_pos++;
	}
      in_pos++;
      return TOK_STRING;
    }

  uint c = *in_pos++;
  switch (c)
    {
    case '=':
      return TOK_EQUAL;
    case ',':
      return TOK_COMMA;
    case ':':
      return TOK_COLON;
    case '(':
      return TOK_OPEN_PAREN;
    case ')':
      return TOK_CLOSE_PAREN;
    case '{':
      return TOK_OPEN_BRACE;
    case '}':
      return TOK_CLOSE_BRACE;
    default:
      if (c == '.' && *in_pos == '.')
	{
	  in_pos++;
	  return TOK_DOTDOT;
	}
      err("Unrecognized character '%c'", c);
    }
}

static token_type next_token()
{
  if (buffered_token != TOK_NONE)
    this_token = buffered_token;
  else
    this_token = get_next_token();
  buffered_token = TOK_NONE;
  return this_token;
}

static void return_token()
{
  assert(this_token != TOK_NONE);
  assert(buffered_token == TOK_NONE);
  buffered_token = this_token;
  this_token = TOK_NONE;
}

static token_type peek_token()
{
  next_token();
  return_token();
  return buffered_token;
}

static bool token_is_int()
{
  return token_num == (double)(int) token_num;
}

/*** Argument types ***/

class arg_int : public arg_val {
  int val;
public:
  arg_int(int x) { val = x; }
  bool given() { return true; }
  int as_int(int def UNUSED) { return val; }
  string dump() { return "<int> " + to_string(val); }
};

class arg_double : public arg_val {
  double val;
public:
  arg_double(double x) { val = x; }
  bool given() { return true; }
  double as_double(double def UNUSED) { return val; }
  string dump() { return "<double> " + to_string(val); }
};

class arg_string : public arg_val {
  string val;
public:
  arg_string(string x) { val = x; }
  bool given() { return true; }
  const string as_string(string def UNUSED) { return val; }
  string dump() { return "<string> \"" + val + '"'; }
};

arg_val null_arg;

/*** Parser ***/

struct unit {
  const char *name;
  double multiplier;
};

#define MM (72/25.4)

static const unit units[] = {
  { "mm",	MM },
  { "cm",	10*MM },
  { "dm",	100*MM },
  { "m",	1000*MM },
  { "in",	72 },
  { "pt",	1 },
  { NULL,	0 }
};

static double parse_dimen(const arg_def *adef)
{
  token_type t = next_token();
  if (t != TOK_NUMBER)
    err("Argument %s must be a dimension", adef->name);
  double tmp = token_num;

  t = next_token();
  if (t != TOK_IDENT)
    {
      if (is_zero(tmp))
	{
	  return_token();
	  return 0;
	}
      err("Argument %s must have a unit", adef->name);
    }
  for (uint i=0; units[i].name; i++)
    if (token == units[i].name)
      return tmp * units[i].multiplier;
  err("Unknown unit %s", token.c_str());
}

static void parse_pipeline(cmd *c)
{
  pipeline *pp = new pipeline;
  next_token();

  while (peek_token() != TOK_CLOSE_BRACE)
    {
      if (pp->branches.size() && next_token() != TOK_COMMA)
	err("Comma expected between pipeline branches");

      pipeline_branch *pb = new pipeline_branch;
      pp->branches.push_back(pb);

      token_type t;
      for (;;)
	{
	  t = next_token();
	  if (t == TOK_END)
	    err("Missing close brace");
	  if (t == TOK_CLOSE_BRACE || t == TOK_COLON || t == TOK_COMMA)
	    break;

	  pipeline_selector ps;
	  if (t != TOK_NUMBER)
	    err("Pipeline selectors must start with a number");
	  if (!token_is_int())
	    err("Pipeline selectors must be integers");
	  ps.from = (int) token_num;
	  ps.to = ps.from;

	  if (peek_token() == TOK_DOTDOT)
	    {
	      next_token();
	      t = next_token();
	      if (t != TOK_NUMBER)
		err("Pipeline selectors must be numbers or ranges");
	      if (!token_is_int())
		err("Pipeline selectors must be integers");
	      ps.to = (int) token_num;
	    }

	  pb->selectors.push_back(ps);
	}

      if (t == TOK_COLON)
	parse_commands(pb->commands);
      else
	return_token();
    }

  c->pipe = pp;
  next_token();
}

static void parse_args(cmd *c)
{
  const cmd_def *cdef = c->def;
  const arg_def *adefs = cdef->arg_defs;
  uint num_args = 0;
  while (adefs[num_args].name)
    num_args++;

  token_type t = next_token();
  if (t != TOK_OPEN_PAREN)
    {
      return_token();
      return;
    }

  bool saw_named = false;
  uint next_pos = 0;
  for (;;)
    {
      t = next_token();
      if (t == TOK_CLOSE_PAREN)
        break;

      while (next_pos < num_args && !(adefs[next_pos].type & AT_POSITIONAL))
	next_pos++;

      uint argi = 0;
      bool has_value = false;

      if (t == TOK_IDENT)
	{
	  while (adefs[argi].name && token != adefs[argi].name)
	    argi++;
	  if (adefs[argi].name)
	    {
	      if (c->args.count(token))
		err("Argument %s given multiple times", token.c_str());
	      t = next_token();
	      if (t == TOK_EQUAL)
		has_value = true;
	      else
		return_token();
	      saw_named = true;
	    }
	  else if (next_pos < num_args && (adefs[next_pos].type & AT_TYPE_MASK) == AT_STRING)
	    {
	      // Shortcut syntax: positional arguments of string type can be specified
	      // as bare identifiers if they do not collide with names or other arguments.
	      return_token();
	      argi = next_pos++;
	      has_value = true;
	    }
	  else
	    err("Command %s has no argument %s", cdef->name, token.c_str());
	}
      else if (saw_named)
	err("Positional arguments must precede named ones");
      else if (next_pos < num_args)
	{
	  return_token();
	  argi = next_pos++;
	  has_value = true;
	}
      else
	err("Too many positional arguments for command %s", cdef->name);

      const arg_def *adef = &adefs[argi];
      uint type = adef->type & AT_TYPE_MASK;
      arg_val *val = NULL;
      if (has_value)
        {
	  switch (type)
	    {
	    case AT_STRING:
	      t = next_token();
	      if (t != TOK_STRING && t != TOK_IDENT)
		err("Argument %s must be a string or identifier", adef->name);
	      val = new arg_string(token);
	      break;
	    case AT_INT:
	      t = next_token();
	      if (t != TOK_NUMBER || !token_is_int())
		err("Argument %s must be an integer", adef->name);
	      val = new arg_int((int) token_num);
	      break;
	    case AT_DOUBLE:
	      t = next_token();
	      if (t != TOK_NUMBER)
		err("Argument %s must be a number", adef->name);
	      val = new arg_double(token_num);
	      break;
	    case AT_DIMEN:
	      val = new arg_double(parse_dimen(adef));
	      break;
	    case AT_SWITCH:
	      t = next_token();
	      if (t != TOK_NUMBER || !token_is_int() || ((int) token_num != 0 && (int) token_num != 1))
		err("Argument %s must be a switch", adef->name);
	      val = new arg_int((int) token_num);
	      break;
	    default:
	      abort();
	    }
	}
      else
        {
	  if (type == AT_SWITCH)
	    val = new arg_int(1);
	  else
	    err("Argument %s must have a value", adef->name);
	}

      c->args[adef->name] = val;

      t = next_token();
      if (t == TOK_CLOSE_PAREN)
	break;
      if (t != TOK_COMMA)
	err("Comma expected after argument %s", adef->name);
    }

  for (uint i=0; i<num_args; i++)
    if ((adefs[i].type & AT_MANDATORY) && !c->args.count(adefs[i].name))
      err("Command %s is missing an argument %s", cdef->name, adefs[i].name);
}

static void debug_cmd(cmd *c, uint indent=0)
{
  printf("%*sCommand %s\n", indent, "", c->def->name);
  for (uint i=0; c->def->arg_defs[i].name; i++)
    {
      const arg_def *adef = &c->def->arg_defs[i];
      string dump = c->arg(adef->name)->dump();
      printf("%*sArg #%d: %s = %s\n", indent+4, "", (int) i, adef->name, dump.c_str());
    }
  if (c->pipe)
    {
      printf("%*sPipeline:\n", indent+4, "");
      for (auto pb: c->pipe->branches)
	{
	  printf("%*sSelector:\n", indent+8, "");
	  for (auto ps: pb->selectors)
	    printf("%*s%d - %d\n", indent+12, "", ps.from, ps.to);
	  printf("%*sCommands:\n", indent+8, "");
	  for (auto cc: pb->commands)
	    debug_cmd(cc, indent+12);
	}
    }
}

static void debug_cmds(list<cmd *> &cmds)
{
  for (auto c: cmds)
    debug_cmd(c);
}

static cmd *parse_cmd()
{
  const cmd_def *cdef = cmd_table;
  while (cdef->name && token != cdef->name)
    cdef++;
  if (!cdef->name)
    err("Unknown command %s", token.c_str());

  cmd *c = new cmd;
  c->def = cdef;
  c->pipe = NULL;

  parse_args(c);

  if (peek_token() == TOK_OPEN_BRACE)
    {
      if (!cdef->has_pipeline)
	err("Command %s does not accept a pipeline", cdef->name);
      parse_pipeline(c);
    }
  else if (cdef->has_pipeline)
    err("Command %s requires a pipeline", cdef->name);

  return c;
}

static void parse_commands(list<cmd *> &cmds)
{
  for (;;)
    {
      token_type t = next_token();
      if (t != TOK_IDENT)
	{
	  return_token();
	  return;
	}

      cmd *c = parse_cmd();
      cmds.push_back(c);
    }
}

static void instantiate(list<cmd *> &cmds)
{
  for (auto c: cmds)
    {
      try
	{
	  c->exec = c->def->constructor(c);
	}
      catch (exception &e)
	{
	  die("Error in %s: %s", c->def->name, e.what());
	}
      if (c->pipe)
	{
	  for (auto pb: c->pipe->branches)
	    instantiate(pb->commands);
	}
    }
}

void parse(const char *in, list<cmd *> &cmds)
{
  debug("### Parsing commands");
  in_pos = in;

  try
    {
      parse_commands(cmds);
      if (next_token() != TOK_END)
	err("Extra tokens after commands");
    }
  catch (exception &e)
    {
      die("Parse error: %s (at position %d)", e.what(), (int)(in_pos - in));
    }

  if (debug_level > 1)
    debug_cmds(cmds);
  instantiate(cmds);
}

void parser_help()
{
  for (int i=0; cmd_table[i].name; i++)
    {
      const cmd_def *def = &cmd_table[i];
      printf("%s - %s\n", def->name, def->help);

      const arg_def *arg = def->arg_defs;
      static const char * const type_names[] = {
	"string", "int", "real", "dimen", "switch"
      };
      while (arg->name)
	{
	  char a[32];
	  snprintf(a, sizeof(a), "%s%s=%s<%s>",
	    (arg->type & AT_POSITIONAL) ? "[" : "",
	    arg->name,
	    (arg->type & AT_POSITIONAL) ? "]" : "",
	    type_names[arg->type & AT_TYPE_MASK]);
	  printf("   %-20s %s\n", a, arg->help);
	  arg++;
	}

	if (def->has_pipeline)
	  printf("   { pipeline }\n");
    }
}
